#if !defined (GEODESIC_BOUNDARY_H)
#define GEODESIC_BOUNDARY_H

#include <complex>

#include "defs.h"
#include "ograph.h"
#include "faces.h"
#include "surfaces.h"

//Vertice con variabili theta
class tvertice:public overtice{ 
 protected:
  long double theta[6];
  mutable long double circo_raggio[4];
  bool cusp[4];  //dice se la parte di bordo corrispondente e' cuspide
 public:
  tvertice();
  tvertice(const overtice& );
  long double get_circo_raggio(int i) const {return circo_raggio[i];}
  void set_circo_raggio(int i, long double r) const {circo_raggio[i]=r;}
  long double get_theta(int i) const {return theta[i];}
  void set_theta(int i, long double th) {theta[i]=th;}
  int get_cusp(int i) const {return cusp[i];}
  void set_cusp(int i, int cp) {cusp[i]=cp;}

  void scrivi_circo_raggio(std::ostream& f) const;
  
  long double c(int i, int j, int a, int b) const;
  long double dc(int i, int j, int a, int b, int r, int s) const;
  long double d(int i, int j, int a) const;
  long double dd(int i, int j, int a, int r, int s) const;
  long double l(int i, int j) const;
  long double dl(int i, int j, int r, int s) const;
  long double cos(int i, int j) const;
  long double sin(int i, int j) const;
	
  bool soluzione_nel_dominio() const;
  std::vector<std::vector<long double> > gram() const;
  double volume() const;
  std::complex<double> U(std::complex<double> z) const;
  double D(int i) const;
  double tilt(int i) const;
  double g() const;
  
  virtual ~tvertice() {}
  virtual tvertice* clone() const {return new tvertice(*this);}
};

//Grafo con variabili theta. I vertici devono essere tvertici.
class tgrafo:public ografo{
  grafo3 bd;
  superficie s;
  bool ha_cuspidi;
 public:
  tgrafo():ografo(), bd(), s() {}
  tgrafo(const tgrafo& o);
  tgrafo(const ografo& o);
  tgrafo(std::istream&, int tipo_formato=0);
  
  virtual ~tgrafo() {}
  tgrafo& operator=(const tgrafo& o);
  virtual tgrafo* clone() const {return new tgrafo(*this);}

  void set_facce();

  long double get_circo_raggio(int i, int j) const 
		{return ((tvertice*) v[i])->get_circo_raggio(j);}
  void set_circo_raggio(int i, int j, long double r) const
		{((tvertice*)v[i])->set_circo_raggio(j,r);}
  long double get_theta(int i, int j) const 
		{return ((tvertice*) v[i])->get_theta(j);}
  void set_theta(int i, int j, long double th) 
		{((tvertice*)v[i])->set_theta(j,th);}
  int get_cusp(int i, int j) const 
		{return ((tvertice*) v[i])->get_cusp(j);}
  void set_cusp(int i, int j, int cp) const
		{((tvertice*)v[i])->set_cusp(j,cp);}
  bool get_ha_cuspidi() const {return ha_cuspidi;}
  
  std::vector<long double> get_F() const;
	std::vector<long double> get_F_boundary() const;
  std::vector<std::vector<long double> > dF() const;
  void reset_theta();           //mette -1 ai theta

  long double somma_theta(int num_faccia) const;
  bool soluzione_geometrica() const;
	bool soluzione_nel_dominio(int num_vertice = -1) const;

  virtual void salva(std::ostream&, int tipo_formato = 0) const;
	virtual void scrivi_geometria(std::ostream & f) const;
  virtual void scrivi(std::ostream& f, int tipo_formato = 0) const;

  void scriviF(std::ostream& f, const std::vector<long double>& F) const;
  void scriviX(std::ostream& f, const std::vector<long double>& X) const;
  void scriviJ(std::ostream& f, const std::vector<std::vector<long double> >& J) const;
  void scrivi_theta(std::ostream& f) const;
  void scrivi_lunghezze_spigoli(std::ostream& fstr) const;

	double volume() const;

  void newton(int ripeti_quante_volte = 5000, std::ostream& = std::cout, 
              bool scrivi_X_F = 0);

	bool canonize();
	bool canonize_aux();

	long double get_l(int num_faccia, int t) const;
  long double get_dl(int vert, int i, int j, int r, int s) const;

	//funzioni di gestione del bordo:
	long double get_lb(int num_vert3, int usc3) const;
	void set_bordo();
	void espandi_raggi(const std::vector<double>& r) const;
	void scrivi_circo_raggio(std::ostream& f) const;
  
	std::vector<double> tilt() const;
	bool canonica() const;
	bool sub_canonica() const;

	bool operator<(const tgrafo&) const;
	
 protected:
  bool newton(bool scrivi_X_F, std::ostream& f);  
};

long double determinant(std::vector<std::vector<long double> >);

std::complex<double> dilog(const std::complex<double>& z);

class invarianti {
 public:
	double vol;
	superficie bordo;
	std::vector<std::vector<long double> > angoli;
	int quante;
	invarianti(double v, const superficie& bd, const std::vector<std::vector<long double> >& ang, int q):
		vol(v), bordo(bd), angoli(ang), quante(q) {}
	invarianti(const tgrafo&);
	bool operator<(const invarianti&) const;
	void ordina();
	void scrivi(bool scrivi_angoli = FALSE) const;
};

bool operator==(const invarianti& inv1, const invarianti& inv2);
void ordina(std::list<invarianti>&);
std::list<invarianti> get_invarianti(const std::list<tgrafo>&);
void newton(std::list<tgrafo>& ls);
bool stessi_angoli(const invarianti& inv1, const invarianti& inv2);
void controllo_volumi(const std::list<tgrafo>& ls);

#endif

